/***************************************************************

This is a data accumulator for the CES system. I'd have used a 
 regular array, but I thought this would be more expandable
 should I make ANOTHER structure change later. 
Its still unknown if this will be the level that has the 
 concurency locks, cause I havn't even thought about
 how this will be threaded.
 
 One of the ideas here is to make this thing so that its internal 
   functionality can be completely overhauled without changing 
   the way that its called by other objects.

 This thing is supposed to store data, and provide for easy 
   manipulation of said data. This SHOULD be able to handle 
   binary data just fine.
   

Methods are sumarized as follows:
  long   GetSize( void );                how big is it?
  long   ReadN(  long position, char * buffer, long size ); read bytes
  long   SetN( long destStart, long destSize, char * src, long srcSize); insert/delete content. automatically resizes accumulator. 
 
  -- internal methods --
  int    ReSize( long newSize);              raw resize op.
  long   Insert( long position, long size);   position specific resize
  long   Remove( long position, long size);   position specific unresize

 Dont use tab characters in this code.
  Dont use tab characters in this code.
   Dont use tab characters in this code.
    Dont use tab characters in this code.
     Dont use tab characters in this code.
      Dont use tab characters in this code.
       Dont use tab characters in this code.
        Dont use tab characters in this code.
         Dont use tab characters in this code.
          thanks.
 
 Creation Dan, <Dec 28/2001>
 added resize. Dan <jan ??/2002>
 added insert and remove. Dan <feb 18/2002>
 added adjustsize. Dan <feb 19/2002>
 removed local class def. Dan <mar 5/2002>
 minor overhaul on size tracking. Dan, <mar 8/2002>
 major overhaul. removed several sizing classes in preference of 
    united content/resize method. remaining methods are private Dan. 
    attempted optimizations of insert / delete code <jul 5/2002>
 rewritten to use block allocation. moorcr0  <2002 july 06>
 added SerialNumber. Dan. <Nov 24 2002>

***************************************************************/

#include <stdlib.h>
#include <string.h>
#include "danStuff.h" //MIN , InLimits
#include "returnValues.h" 
#include "accumulator.h" // class definition

#include <stdio.h> //for printf

//#include <always_exclude.h>


/*****************************************
*
* Initilize contents
*  By default the buffer has one byte
*
*****************************************/
Status_t  TACInit   ( tac_t * this){
  this->blockCount   = 1;
  this->contents     = malloc(BLOCK*(this->blockCount)); //new char[BLOCK*(this->blockCount)];
  this->contentSize  = 1;
  this->SerialNumber = 0;
  if (this->contents != NULL) {
    *this->contents = 0;
    return OK;
  } else {
    return  ErrMalloc;
  }
}


/****************************************
 *
 *  standard deconstructor
 *
 ***************************************/
Status_t  TACFini ( tac_t * this) {
 free(this->contents);//  delete[] contents;
 this->contents    = NULL;
 this->blockCount  = 0;
 this->contentSize = 0;
 this->SerialNumber++;
 return OK;
}


/********************************************************************
*
* Read N bytes of the accumulator. The read 
*  and write methods exist to enable 
*  this code to be concurrent.
*  
*  In:
*    position  :  The First character to start reading from (included)
*    buffer    :  The location of a buffer to store the result.
*    size      :  The number of characters to read.]
*    returnSize:  The Number of characters actually read. (This may 
*     come up short if the buffers end was reached.)
*
* notes:
*
*   - should this be able to handle negitive sizes?
*
*******************************************************************/
Status_t TACReadN ( tac_t * this, AOffset_t position, char * buffer, AOffset_t size, AOffset_t * returnSize ){

  AOffset_t safeSize;
  
  safeSize = Min(position + size, this->contentSize) - position; 
  
  //safety check.
  if (( position > this->contentSize )|( position < 0 )) { 
    return BadArg;
  }   
  
  // I dont want to limit this to strings, hence memcpy
  memcpy(buffer, (this->contents)+position, safeSize);

  if (returnSize)  *returnSize = safeSize-1; // -1 cause there is a terminating null that technically doesn't exist.

  return OK;
}








Status_t TACCountOccurances ( tac_t * this, char c, AOffset_t * count ){

  AOffset_t p;
  
  *count = 0;
  
  for (p = 0; p < this->contentSize; ( *((this->contents)+(p++)) == c)?((*count)++):0);

  return OK;
}


/*************************************************************************
*
* Insert/delete content. automatically resizes accumulator.
*  
* In:
*   destStart: First position of the accumulators contents to replace.
*   destSize:  Number of accumulator "bytes" to be affected by "paste".
*   src:       Data to use for the contents of the "paste".
*   srcSize:   Size of src.
*
* Out:
*   The number of characters written, should match srcSize if there
*      were no problems
*
**************************************************************************/
Status_t  TACSetN ( tac_t * this, AOffset_t destStart, AOffset_t destSize, char * src, AOffset_t srcSize ){

   AOffset_t memDelta;

   //sanity check, its 1:30 am, I must not be sane...
   if ( !WITHIN(0, this->contentSize, destStart) ) {
     return BadArg;
   }  

   if ((destSize < 0) | (srcSize < 0)) {
     return BadArg;
   }
   //sanity check, you can only replace content up to the end
   //  of the buffer, but not past it

   destSize = Min((this->contentSize)-destStart, destSize); 

   // sanity check, only YOU can prevent buffer overuns.

   memDelta = srcSize - destSize ;

   //printf("accumulator: delta size is %ld\n", memDelta);
 
   if(memDelta >0)   {
      if ( (this->contentSize) + memDelta > (this->blockCount) *BLOCK)      {
        //we need memory
        TACReSize( this,  (((this->contentSize) + memDelta) /BLOCK)+1);
      }
      TACshift(this, destStart, memDelta);
   }   else if (memDelta <0)   {
      TACshift(this, destStart, memDelta);
      if((this->contentSize) +memDelta < ((this->blockCount)/2 *BLOCK))      {
         TACReSize(this, (this->blockCount)/2);
      }
   }
   //printf("accumulator: resize done, copying contents\n");
   //printf("accumulator: contents+destStart %x, src %s, srcSize %ld \n", 
   //(contents+destStart), src, srcSize);
   memcpy( (this->contents)+destStart, src, srcSize);    // write content
   this->SerialNumber++;
   return OK;
}


/****************************************************
 *
 *  Resize the accumulator
 * This will resize that array to specified size
 *  @param   newBlockSize  is the new buffer size in blocks
 *  @return  1 success    0 bad
 *
 ***************************************************/
Status_t TACReSize  ( tac_t * this, AOffset_t newBlockSize ){
  char * temp ;
  
  temp = malloc(newBlockSize * BLOCK); //!!!???!!! check for errors *&^&^%IT!
  memcpy(temp, this->contents, this->contentSize);
  free(this->contents); 
  this->contents   = temp;
  this->blockCount = newBlockSize;
  return OK;
}


/*****************************************************
 * 
 *  Shift contents of the array, will need to be changed later
 *    since O(n) time here.
 *    
 *  IN: 
 *   offset: the first character to be effected
 *   delta:  the number of characters after offset to 
 *      be effected. a negitive number 'removes' delta,
 *      a positive number 'adds' delta
 *
 *  OUT:
 *   dosn't mean anything yet, ignore
 *
 *****************************************************/
Status_t TACshift  ( tac_t * this, AOffset_t offset , AOffset_t delta )   {

  if (delta ==0)
    return OK;
  
  if (delta < 0)  { //contracting (shrink)
    char * src  = (this->contents) + offset + (- delta);
    char * dest = (this->contents) + offset;
    
    // fold back the post data
    for (; src <= ((this->contents) + (this->contentSize)); src++) {
      *dest = *src;
      dest++;
    }
    
  } else  {  //expanding
    
    char * dest = this->contents + (this->contentSize) + delta; 
    //ptr math -> to end of buffer
    char * src  = dest - delta;// delta back from end of buffer
    
    for (; src >= ((this->contents) + offset);src--)  {
        *dest = *src;
        *dest--;     
      }
  }
  
  this->contentSize+=delta;
  this->SerialNumber++;
  return OK;

}


/*
int main(void) {

  tac_t A;
  char foo[400];
  
  TACInit(&A);
  
  TACSetN ( &A, 0, 0, "world.", 6 );
  TACReadN ( &A, 0, foo, 20, NULL );
  printf("%s\n", foo);
  
  TACSetN ( &A, 0, 0, "Hello ", 6 );  
  TACReadN ( &A, 0, foo, 20, NULL );
  printf("%s\n", foo);
  
  TACSetN ( &A, 6, 6, "Rue.", 4 );  
  TACReadN ( &A, 0, foo, 20, NULL );
  printf("%s\n", foo);
  
  TACFini(&A);

  return 0;
}
**/


