/*

https://viewsourcecode.org/snaptoken/kilo/04.aTextViewer.html

step 149

bugs:

  - rainbow brackets can be a bit.... dynamic.....
  - conflict between cursor position during search
  - mouse wheel scroll is bound to keeping cursor on screen
  
Features in queue:

  - block hilighting
    proper hilighting for comment blocks, strings
  - copy, cut and paste
    the extravogance!
  - undo
    wtf, you trying to kill me?
  - replace (as in search/replace)
  - split view
    oooh CMON NOW....
  - block indent
  - rectangular select
  - mouse on/off
  
  
Directions:
  - rename terminal functions
  - use an event loop.  
  
Manual:

  - compile with gcc *.c
  - To get the number of a terminal run:    tty
  - open a spare terminal and get its tty  (for debug stream)
  - in a deifferent terminal run with:  ./a.out foo.txt  2>/dev/pts/13
    (or whatever terminal number you got)
    

  


  ctrl-q   quit
  ctrl-s   save
  ctrl-f   find
  ctrl-n   new
  ctrl-o   open   
  
  alt-0    insert main() function
  alt-1    insert an if block
  alt-2    insert an else blocl
  alt-3    insert a for loop
  alt-4    insert a while loop  
  
  F3       find
    (in find)
    F3 find next
    shift-F3 find previous
    esc    abort
    enter  exit find with cursor position on found.
 

*/

#define _GNU_SOURCE     // asprintf

#include <stdio.h>
#include <unistd.h>
#include <stdint.h>
//#include <termios.h>    // raw mode
#include <stdlib.h>     // raw mode
//#include <ctype.h>      // keypress
//#include <errno.h>      // die
//#include <string.h>     // writes
//#include <sys/ioctl.h>  // getwindowsize
#include <sys/types.h>  // file openingg
#include <time.h>       // status message
#include <stdarg.h>     // status message
#include <fcntl.h>      // save
#include <stropts.h>

#include "danStuff.h"
#include "buffedit.h"
#include "terminal.h"
#include "main.h"

#include "accumulator.h"
#include "filefilter.h"
#include "navigator.h"



substTable_t    SUBLIST[]    = { {'\t', "  "}, /*{'h', "hello_world"},   {'r', "Rue mohr" },*/ { 0,""} };  // I officially hate tabs more now.

char * KEYWORDS       = " switch case default if else do while for break continue return ";
char * DIRECTIVES     = " #include #define #ifdef #ifndef #endif ";
char * TYPES          = " int char unsigned float double long uint8_t uint16_t uint32_t int8_t int16_t int32_t const enum struct union static typedef void ";
char * SEPCHARS       = " \t,.+-/*=~%&|^!?()<>[]{}'\";:";
char * INSERT_MAIN[]  = { "#include <stdio.h>", "  ", "int main(int argc, char *argv[], char *envp[]) { ", "  ", "  return 0; ", "}",  "" };
char * INSERT_IF[]    = { "if (  ) {", "  ", "}",                                            "" };
char * INSERT_ELSE[]  = { "else {", "  ", "}",                                               "" };
char * INSERT_FOR[]   = { "for ( ; ; ) {", "  ", "}",                                        "" };
char * INSERT_WHILE[] = { "while( ) {", "  ", "}",                                           "" };







int main(int argc, char *argv[]) {
    
  Editor_t  E;  
  int c;
    
  enableRawMode() ;
  initEditor(&E);
  TACInit( &(E.accum) );
  
  if (argc >= 2)  editorOpen( &E, argv[1] );
  
  editorSetStatusMessage(&E, "Quit: Ctrl-Q, Save: Ctrl-S, Find: F3, New: Ctrl-N, Open: Ctrl-O ");
  
  while (1) {
  
     editorRefreshScreen(&E);
    
     c = checkStdin();
     if (0) {
     } else if (E.focus == FOCUS_DOCUMENT) {
      editorProcessEvent(&E, c);
     } else if (E.focus == FOCUS_PROMPT) {
      promptProcessEvent(&E, c) ;
     }  
     
  //   if (E->tmpMessage) 
  //     if (time(NULL) > E->messageExpire) return TIMER_EVENT;         // trip out for status bar to reset.
     
  }
  
  return 0;
}

// -----------------------------------------------------------------



/*** input ***/
void editorProcessEvent(Editor_t * E, int c) {
  int times;
  
  switch (c) {
  
    case KEY_NONE:
    case TIMER_EVENT:  // used to trip screen redraw
    break;
  
   // commands:
    case CTRL_A:
      editorCmdSaveAs(E);
    break;
   
    case CTRL_S:       // save
      editorCmdSave(E);
    break;

    case CTRL_N:        // new
      editorNew(E);
    break;

    case CTRL_O:        // open
      editorCmdOpen(E);
    break;
  
    case CTRL_F:         // find
    case KEY_F3:
      editorFindCallback(E, FIND_RESET); // reset the callback for find.      
      editorFind(E);
    break;
  
    case CTRL_Q:       // quit
      writes(ANSI_Cls);
      writes(ANSI_Home);
      exit(0);
    break;
    
    case ALT_0:        // insert code block
    case ALT_1:
    case ALT_2:
    case ALT_3:
    case ALT_4:
      editorInsertFlow(E, c);
    break;    
    
    
      
   // navigation
    case KEY_HOME:  E->crs_curr.cx = 0;                  break;  
      
    case KEY_END: /*  E->crs_curr.cx = strlen(E->row[E->crs_curr.cy].chars);*/   break;   
     
    case PAGE_UP:
    case PAGE_DOWN:      
        for ( times = E->windowrows; times--; )
          editorMoveCursor(E, c == PAGE_UP ? KEY_UP : KEY_DOWN);      
    break;    
    
    case KEY_LEFT:
    case KEY_RIGHT:
    case KEY_UP:
    case KEY_DOWN:
      editorMoveCursor(E, c);
    break;
    
    case MOUSE_EVENT:
      fprintf(stderr, "Mouse Event: button: %X X:%d Y:%d\n", M.buttons, M.X, M.Y);
    /*
      if (0) {
      } else if (M.buttons == 0x60) { if (E->crs_curr.rowoff) E->crs_curr.rowoff--; // editorMoveCursor(E, KEY_UP); // scroll up
      } else if (M.buttons == 0x61) { E->crs_curr.rowoff++; // editorMoveCursor(E, KEY_DOWN); // scroll down
      }*/
    break;
      
            
   // entry
    case '\r':
     editorInsertChar(E,'\n');
     E->crs_curr.cx = 0;
     E->crs_curr.cy++;
     E->lineCount++;
    break;
    
    case '\b':
    case BS_KEY:  // backspace is actually a bastard of an operation!
      if ((E->crs_curr.cx == 0) & (E->crs_curr.cy == 0)) break; // your at the start of the file, NO.
      editorBackspaceChar( E );     
    break;  
       
    case KEY_DEL:  // The accumulator catches insane requests, such as delete beyond end of content
      editorDelChar(E);      
    break;                            
    
    default:
    
      if   (isprint(c) || (c == 9)) editorInsertChar(E,c);          
      else fprintf(stderr, "hu? -> %X\n", c);  
      
    break;
        
  }
  
}

void editorBackspaceChar( Editor_t * E ){
    char MT[1];
    Nav_t TNav;     
    AOffset_t crs;  
    AOffset_t len; 
    
    NavInit ( &TNav );
    NavSetAccumulator ( &TNav, &(E->accum));
  
    if (E->crs_curr.cy <= E->lineCount)   {  // are we off the end of the file?
      crs = NavFwdLines (&TNav, 0, E->crs_curr.cy);  
      if ((len = NavLineLength (&TNav, crs)) >= E->crs_curr.cx)   {  // are we off the end of the line content?
        crs += E->crs_curr.cx;
        TACSetN ( &(E->accum), crs-1, 1, MT, 0);                
      }
    }
        
    if (E->crs_curr.cx > 0) { // move the cursor back a content position.
      E->crs_curr.cx--;
    } else {
      E->crs_curr.cy-- ;     
      E->crs_curr.cx =  NavLinePos  ( &TNav, crs-2) + 1;
    }


}

void editorDelChar ( Editor_t * E ){
    char MT[1];
    Nav_t TNav;     
    AOffset_t crs;  
    AOffset_t len; 
    
    NavInit ( &TNav );
    NavSetAccumulator ( &TNav, &(E->accum));
  
    if (E->crs_curr.cy > E->lineCount) return;  // beyond end of file.
    crs = NavFwdLines (&TNav, 0, E->crs_curr.cy);  
    if ((len = NavLineLength (&TNav, crs)) < E->crs_curr.cx) { // beyond end of line.
      // oooh boy, ok, so we want to pad the existing line to this location, then append the previous line to it.
      len = E->crs_curr.cx - len;
      
    } else { 
      crs += E->crs_curr.cx;    
      TACSetN ( &(E->accum), crs, 1, MT, 0);
    }
}



void editorInsertChar(Editor_t * E,int c) {

  Nav_t TNav;     
  AOffset_t crs;  
  AOffset_t len; 
  char t[] = "\0\0";
  char *pad;
  
  // map cursor to file, it may not even be within the files content
    // find the line were on
  NavInit ( &TNav );
  NavSetAccumulator ( &TNav, &(E->accum));
  t[0] = c;
  
  if (E->crs_curr.cy >= E->lineCount) {  // add lines to the file as required.
    fprintf(stderr, "adding %d lines\n", E->crs_curr.cy - E->lineCount + 1);
    pad = aGenRun( '\n', E->crs_curr.cy - E->lineCount + 1);
    fprintf(stderr, "TAC: %d bytes, ins @ %d\n", E->accum.contentSize, (E->accum.contentSize) - 1);
    TACSetN ( &(E->accum), (E->accum.contentSize) - 1, 0, pad, E->crs_curr.cy - E->lineCount + 1 );   
    free(pad);
    
    TACCountOccurances ( &(E->accum), '\n', &(E->lineCount) );
    E->lineCount++;
  } 
  crs = NavFwdLines (&TNav, 0, E->crs_curr.cy);
  
  if ((len = NavLineLength (&TNav, crs)) < E->crs_curr.cx) { // make the line longer as required
    pad = aGenRun( ' ', E->crs_curr.cx - len + 1);
    
    fprintf(stderr, "TAC: %d bytes, ins @ %d\n", E->accum.contentSize, NavLineEnd (&TNav, crs) );
    
    TACSetN ( &(E->accum), Min(NavLineEnd (&TNav, crs), E->accum.contentSize - 1)  , 0, pad, E->crs_curr.cx - len + 1);   
    free(pad);
  }      
  crs += E->crs_curr.cx;
  
  // insert char.
  fprintf(stderr, "TAC: %d bytes, ins @ %d\n", E->accum.contentSize, crs);
  TACSetN ( &(E->accum), crs, 0, t, 1 );
  
  E->crs_curr.cx++;
  

}





void editorInsertFlow(Editor_t * E, int c) {
/*
   char * line;
   char * t;
   char * pad;
   int i;
   
   pad = calloc(1, E->crs_curr.cx+1); memset(pad, ' ', E->crs_curr.cx);      
   
   switch (c) {
    case ALT_0:  // main loop      
      for(i = 0, line = INSERT_MAIN[0]; *(line = INSERT_MAIN[i]) ; i++) {
        editorInsertRow(E, E->crs_curr.cy+i, line, strlen(line)); 
      }
      E->crs_curr.cy+=3;
      E->crs_curr.cx += 2;
    break;
   
    case ALT_1: // if block   
      editorRowInsertString(E, &E->row[E->crs_curr.cy], E->crs_curr.cx, INSERT_IF[0]);
      for(i = 1, line = INSERT_IF[1]; *(line = INSERT_IF[i]) ; i++) {
        asprintf(&t, "%s%s", pad, line);
        editorInsertRow(E, E->crs_curr.cy+i, t, strlen(t)); 
        free(t);
      }
      E->crs_curr.cx += 5;
    break;
    
    case ALT_2: // else block   
      editorRowInsertString(E, &E->row[E->crs_curr.cy], E->crs_curr.cx, INSERT_ELSE[0]); 
      for(i = 1, line = INSERT_ELSE[1]; *(line = INSERT_ELSE[i]) ; i++) {
        asprintf(&t, "%s%s", pad, line);
        editorInsertRow(E, E->crs_curr.cy+i, t, strlen(t)); 
        free(t);
      }
      E->crs_curr.cy++ ;
      E->crs_curr.cx += 2;   
      
    break;
    
    case ALT_3: // for block
      editorRowInsertString(E, &E->row[E->crs_curr.cy], E->crs_curr.cx, INSERT_FOR[0]);
      for(i = 1, line = INSERT_FOR[1]; *(line = INSERT_FOR[i]) ; i++) {
        asprintf(&t, "%s%s", pad, line);
        editorInsertRow(E, E->crs_curr.cy+i, t, strlen(t)); 
        free(t);
      }
      E->crs_curr.cx += 5;   
        
    break;
    
    case ALT_4: // while block
      editorRowInsertString(E, &E->row[E->crs_curr.cy], E->crs_curr.cx, INSERT_WHILE[0]);
      for(i = 1, line = INSERT_WHILE[1]; *(line = INSERT_WHILE[i]) ; i++) {
        asprintf(&t, "%s%s", pad, line);
        editorInsertRow(E, E->crs_curr.cy+i, t, strlen(t)); 
        free(t);
      }
      E->crs_curr.cx += 6; 
      
    break;
   
   }
   
   free(pad);
*/
}





//------------------------------------------------| FIND STUFF |-------------------------------------------------------
/*

repeat F3 for next find
up/down arrows for search history
shift-F3 for reverse find

*/
void editorFind( Editor_t * E ) {
  
  // save orig position.
 // E->crs_bkp.cx = E->crs_curr.cx; E->crs_bkp.cy = E->crs_curr.cy; E->crs_bkp.coloff = E->crs_curr.coloff; E->crs_bkp.rowoff = E->crs_curr.rowoff;
  editorPrompt(E, "Search (F3 Next/shift-F3 Back/Esc/Enter): ", "word", "", editorFindCallback );

}

void editorFindCallback(Editor_t * E, int c) {
/*
  int i;
  char *match;
  erow *row;
  
  static int    n_match = 1;  // if you have more than 32768 matches, I feel for you
         int    hit;

  if (0) {
  } else if (c == SUBMIT_EVENT) {
    E->messageExpire = time(NULL)-1; // get bar to reset
    return;
    
  } else if (c == CANCEL_EVENT) {
    E->crs_curr.cx = E->crs_bkp.cx; E->crs_curr.cy = E->crs_bkp.cy; E->crs_curr.coloff = E->crs_bkp.coloff; E->crs_curr.rowoff = E->crs_bkp.rowoff;    
    E->messageExpire = time(NULL)-1; // get bar to reset
    return;
    
  } else if ( c == KEY_F3 ) {       // find next
    n_match++;
    
  } else if ( c == FIND_RESET ) {
    n_match = 1;
    return;
    
  } else if ( c == KEY_SHIFT_F3 ) { // find prev
    if (n_match > 1) n_match--;
    else {                               // beep to say we cant go earlier than the first match
      putchar( '\a'); fflush(stdout);
      return;
    }
  } else if (!isprint(c)) {         // ignore control stuff
    return;
  }

  for (i = 0, hit = 0; i < E->lineCount; i++) {
    row = &E->row[i];
    match = strstr(row->chars, E->dialog.el.buff); // why are we using render instead of chars???
    if (match) {
      hit++;
      if (hit == n_match) {
        E->crs_curr.cy = i;                                   // we position cx,cy,  editorUpdateRow()  adjusts cx into rx, which is where the cursor is actually put.
        E->crs_curr.cx = match - row->chars + strlen(E->dialog.el.buff) ; // pointers of offset.
        E->crs_curr.rowoff = E->lineCount;
        break;
      }
    }       
  }
  
  if (!(i < E->lineCount)) {  // if no match was found, beep!    
    if (n_match > 1) n_match--; // last rematch didn't work    
    putchar( '\a'); fflush(stdout);
  }
  */
}

//-----------------------------------------------| SAVE |----------------------------------------------

void editorSaveCallback(Editor_t * E, int c) {
  char * filename;

  if (0) {
  } else if (c == SUBMIT_EVENT) {
    filename = strdup(E->dialog.el.buff);    
    editorSave( E, filename ) ; // get bar to reset if open was ok.        
    return;    
  } else if (c == CANCEL_EVENT) {
    editorSetStatusMessage(E, "Canceled.");
    return;    
  } 

}


void editorCmdSaveAs(Editor_t * E) {  
   editorPrompt(E, "Save as: ", "filename.ext", "untitled.txt", editorSaveCallback);
   return; 
}

void editorCmdSave(Editor_t * E) {  
  char *buf ;
  int   fd  ;
                    
  if (E->filename == NULL)  {                                                           // ask for a name if there wasn't one.
    editorPrompt(E, "Save as: ", "filename.ext", "untitled.txt", editorSaveCallback);
    return;    
  } else {
    if (editorSave( E, E->filename ) == 0)  {
      editorSetStatusMessage(E, "Saved.");
      E->dirty = 0;
    } else {
      editorSetStatusMessage(E, "Unable to save %s", strerror(errno));
    }  
    
  }  
}

int editorSave(Editor_t * E, char *filename) {

  if (FF1SaveFile( &(E->accum), filename ) != OK){
    return 1;
  }
  
  free(E->filename);
  E->filename = strdup(filename);
  
  return 0;

}



// ------------------------------------------------| LOAD/OPEN |-----------------------------------------------

// callback is used by the prompt machine when keys come in. Just Submit or Cancel here.
void editorOpenCallback(Editor_t * E, int c) {
  char * filename;

  if (0) {
  } else if (c == SUBMIT_EVENT) {
 
    filename = strdup(E->dialog.el.buff);    
    editorNew(E);
    if (editorOpen( E, filename ) == 0)  E->messageExpire = time(NULL)-1; // get bar to reset if open was ok.        
    return;
  } else if (c == CANCEL_EVENT) {
    editorSetStatusMessage(E, "Canceled.");
    return;    
  } 

}

// prep the prompt and focus us on it.
void editorCmdOpen ( Editor_t * E ) {
 editorPrompt(E, "Open: ", "filename.ext", "", editorOpenCallback); 
}


int editorOpen(Editor_t * E, char *filename) {

  if (FF1LoadFile( &(E->accum), filename ) != OK){
    editorSetStatusMessage(E, "Unable to open file.");
    return 1;
  }
  
  TACCountOccurances ( &(E->accum), '\n', &(E->lineCount) );
  E->lineCount++;
  
  E->dirty = 0;
  free(E->filename);
  E->filename = strdup(filename);
  //fprintf(stderr, "%s", E->filename);
  
  return 0;

}


//--------------------------------------------------------------------------------------------------------------------





void editorNew ( Editor_t * E ) {

  free(E->filename); E->filename = NULL;
  E->crs_curr.cx     = 0, E->crs_curr.cy     = 0;
  E->crs_curr.coloff = 0; E->crs_curr.rowoff = 0;
  TACFini ( &(E->accum) );
  TACInit ( &(E->accum) );
  E->lineCount = 1;
  
}






void initEditor(Editor_t * E) {

  if (getWindowSize(&C.screenrows, &C.screencols) == -1) die("getWindowSize");

  E->filename               = NULL;
  E->windowrows             = C.screenrows-1;
  E->windowcols             = C.screencols;
  E->crs_curr.cx            = 0;
  E->crs_curr.cy            = 0;
  E->rx                     = 0;
  E->crs_curr.rowoff        = 0;
  E->dirty                  = 0;
  E->crs_curr.coloff        = 0;
  E->lineCount              = 1;
  E->focus                  = FOCUS_DOCUMENT;
  
  E->tmpMessage             = NULL;
  E->messageExpire          = 0;  
  E->dialog.prompt          = NULL;
  E->dialog.hint            = NULL;
  buffEdInit(&E->dialog.el);  
      
}



//========================================| Screen drawing stuff |==========================================





void editorRefreshScreen(Editor_t * E) {  

   char * ab = NULL;  // screen buff
   char * temp = NULL;
   
   if (getWindowSize(&C.screenrows, &C.screencols) != -1) {
     E->windowrows    = C.screenrows-1;
     E->windowcols    = C.screencols;
   }
  
   editorScroll(E);
   
   astrcat(&ab,  ANSI_Hide);
   astrcat(&ab,  ANSI_Home);
   astrcat(&ab,  ANSI_ColorRST);
   
   if (time(NULL) > E->messageExpire) {   
     editorDrawStatusBar ( E, &ab);  
     if (E->tmpMessage) {
       free(E->tmpMessage);
       E->tmpMessage = NULL;
     }
   } else {
     editorDrawMessageBar( E, &ab);
   }
   astrcat(&ab, "\r\n");
   editorDrawRows      ( E, &ab);
   
   if (E->focus == FOCUS_DOCUMENT) {
     asprintf(&temp,"\e[%d;%dH", E->crs_curr.cy - E->crs_curr.rowoff + 2, E->rx - E->crs_curr.coloff + 1); // set cursor position
   } else {
     asprintf(&temp,"\e[%d;%dH", 1, E->dialog.el.cursor + strlen(E->dialog.prompt) + 1); // set cursor position
   }   
   
   astrcat(&ab,  temp);
   free(temp);
   
   astrcat(&ab,  ANSI_Show); // show cursor command
   
   writes(ab);  // write to screen
   free(ab);
}





void editorDrawRows(Editor_t * E, char ** ab) {
  int y;
  Nav_t TNav;
  AOffset_t crs;
  char buff[1024]; // !!!???!!! fix this.
  unsigned int len;
  
  NavInit ( &TNav );
  NavSetAccumulator ( &TNav, &(E->accum));
  
  crs = NavFwdLines (&TNav, 0, E->crs_curr.rowoff);
  astrcat(ab, ANSI_INV);
  for (y = 0; y < (E->windowrows) ; y++) {
  
  if ( y + E->crs_curr.rowoff >= E->lineCount) { // past end of file.
    astrcat(ab, "~");
  } else {
    if ((len = NavLineLength ( &TNav, crs)) >  E->crs_curr.coloff  ) { // past end of line.
      TACReadN ( &(E->accum), crs, buff, len , NULL );
      buff[len] = 0;
      astrncat(ab, buff + E->crs_curr.coloff, E->windowcols  );
    }
    crs = NavNextLine ( &TNav, crs);
  }
  
  
  /*
    filerow = y + E->crs_curr.rowoff;
    if (filerow >= E->lineCount) { // past end of file.
      astrcat(ab, "~");
    } else {
      if (E->crs_curr.coloff < strlen(E->row[filerow].render)) {   //  normal content        
       
         for (p = 0; E->row[filerow].render[p+E->crs_curr.coloff] && (p <(( E->windowcols) )) ; p++ ) {  //!!!???!!! ARG per character render
           if (oa != E->row[filerow].attrib[p+E->crs_curr.coloff]) {
             oa = E->row[filerow].attrib[p+E->crs_curr.coloff];
             asprintf(&b, "%s%c", editorAttribToESC( oa ), E->row[filerow].render[p+E->crs_curr.coloff]);
           } else {
             asprintf(&b, "%c",  E->row[filerow].render[p+E->crs_curr.coloff]); 
           }
           astrcat(ab, b);
           free(b);                        
         }
                                        
      } else {                                            // past end of horizontal content
        astrcat(ab, "");        
      }      
      
    }*/
  //  astrcat(ab, "~");
    
    
    astrcat(ab, ANSI_ClrLine);
    if (y < E->windowrows -1)  astrcat(ab, "\r\n");
  }
  
   astrcat(ab, ANSI_NORM);
}



// translate chars buffer to render buffer. then update the hilighting
void editorUpdateRow(Editor_t * E, erow *row) {
}


// ------------------------------------------------------- status bar stuff ------------------------------------------------


void editorSetStatusMessage(Editor_t * E, const char *fmt, ...) {
  va_list ap;
  va_start(ap, fmt);
  
  free(E->tmpMessage);
  asprintf(&E->tmpMessage, fmt, ap);
  
  va_end(ap);
  E->messageExpire = time(NULL) + 3; // you have 3 seconds to realize the clock is about to roll over.
  
}

void editorDrawStatusBar(Editor_t * E, char **ab) {
  char * fill;
  int l;
  char *info1, *info2;
  int srcStart, destStart, len;
  
  astrcat(ab, ANSI_ColorRST);
  astrcat(ab, ANSI_INV);
  
  fill = malloc(E->windowcols+1);
  memset(fill, ' ', E->windowcols); fill[E->windowcols] = 0;
  
  l = asprintf(&info1, "%d,%d", E->crs_curr.cx + 1, E->crs_curr.cy + 1); 
  memcpy(fill, info1, l);
  free(info1);

    asprintf(&info2, "%c%.20s - %d lines - %d bytes",  E->dirty?'*':' ', E->filename ? E->filename : "Untitled", E->lineCount, E->accum.contentSize );    
    l = asprintf(&info1, "%.*s", E->windowcols-1, info2);

    free(info2);

    destStart = Max(0, E->windowcols - 1 - l);
    len       = Min(l ,E->windowcols - 1) ;
    srcStart  = l - len;

    memcpy(&fill[destStart], &info1[srcStart], len);
    free(info1);
      
  astrcat(ab, fill);
  free(fill);

  astrcat(ab, ANSI_NORM);
}


void editorDrawMessageBar(Editor_t * E, char **ab) {
  char * tmp;
  astrcat(ab, ANSI_ClrLine);
  asprintf(&tmp, "%.*s", E->windowcols, E->tmpMessage);
  astrcat(ab, tmp);
  free(tmp);
}



// ------------------------------------------------------- prompt stuff ----------------------------------------------------




void promptProcessEvent(Editor_t * E, int c ) {
  char * rend = NULL;  

    
    c =  promptProcessKeypress(E, &E->dialog.el, c);         
   
         // render status line
    astrcat  ( &rend,  E->dialog.prompt);
    astrcat  ( &rend,  E->dialog.el.buff);    
    if ((strlen(E->dialog.hint) - strlen(E->dialog.el.buff)) > 0) {
      astrcat ( &rend, ANSI_GreyF );
      astrcat ( &rend, &E->dialog.hint[strlen(E->dialog.el.buff)]);
      astrcat ( &rend, ANSI_ColorRST );
    }
    
    editorSetStatusMessage(E, rend);  //
    free(rend); rend = NULL;
    
    
    switch  ( c ) {   
      case   1:             // content not changed, ignore.      
      break;                   
      
      case   0:     // result event
        if (E->dialog.EV_CB) E->dialog.EV_CB(E, SUBMIT_EVENT);
        E->focus = FOCUS_DOCUMENT;
      break;                
      
      case  -1:          // cancel event   
        if (E->dialog.EV_CB) E->dialog.EV_CB(E, CANCEL_EVENT);
        E->focus = FOCUS_DOCUMENT;
      break;                   
      
      default : if (E->dialog.EV_CB) E->dialog.EV_CB(E, c);  // type   event      
    }    
    
   
}




// this stitches our events to the methods for BuffEd
//  return 1 means continue, 0 means normal result, -1 means cancel
int promptProcessKeypress( Editor_t * E, buffEdProc_t * this, int c) {
//  int c; 
  
 // switch (c = checkStdin(E)) {
  switch (c) {
    
    case '\r':
    case '\n':
      return 0; // return valid completion
    break;
    
   // case CTRL_KEY('l'):
    case CTRL_L:
    case '\e':
      return -1; // return abort
    break;
      
    case '\b':
    case BS_KEY:
      buffEdTakeStroke(  this , EV_BACKSP,  ' ');   return 1;
    break;
    
    case KEY_DEL:
      buffEdTakeStroke(  this , EV_DELETE,  ' ');    return 1;  
    break;      
    
    case KEY_HOME: 
      buffEdTakeStroke( this , EV_CRS_START,  ' ');  return 1;
    break;  
      
    case KEY_END: 
      buffEdTakeStroke(  this , EV_CRS_END,  ' ');   return 1;
    break;   
           
    case KEY_LEFT:
      buffEdTakeStroke(  this , EV_CRS_BCK,  ' ');   return 1;
    break;
    
    case KEY_RIGHT:
      buffEdTakeStroke(  this , EV_CRS_FWD,  ' ');   return 1;
    break;
    
    default:
      if (isprint(c)) {
        buffEdTakeStroke( this , EV_INSERT,  c);
      } else {
        fprintf(stderr, "pass -> %04X\n", c);  
      }
      return c;
    break;
     
  }
  
}

// set up prompt mode.
char *editorPrompt(Editor_t * E,  char *prompt, char *hint, char *dfValue, void (*callback)(Editor_t * E, int event)) {

  char * rend = NULL;        
  int c;
                
  free(E->dialog.el.buff); E->dialog.el.buff = strdup(dfValue);
  buffEdTakeStroke( &E->dialog.el , EV_CRS_END,  ' ');  // place cursor at end of default entry  
  free(E->dialog.prompt);  E->dialog.prompt = strdup(prompt);
  free(E->dialog.hint);    E->dialog.hint   = strdup(hint);
  E->dialog.EV_CB = callback;                           // set up event callback
    
  // render status line
  astrcat  ( &rend,  E->dialog.prompt);
  astrcat  ( &rend,  E->dialog.el.buff);    
  if ((strlen(E->dialog.hint) - strlen(E->dialog.el.buff)) > 0) {
    astrcat ( &rend, ANSI_GreyF );
    astrcat ( &rend, &E->dialog.hint[strlen(E->dialog.el.buff)]);
    astrcat ( &rend, ANSI_ColorRST );
  }

  editorSetStatusMessage(E, rend);  //
  free(rend); rend = NULL;

  E->focus = FOCUS_PROMPT;

}



//====================================================================================================


void editorMoveCursor(Editor_t * E, int c) {
  switch (c) {
    case KEY_LEFT:   if (E->crs_curr.cx != 0)    E->crs_curr.cx--;   break;
    case KEY_RIGHT:                              E->crs_curr.cx++;   break;
    case KEY_UP:     if (E->crs_curr.cy != 0)    E->crs_curr.cy--;   break;
    case KEY_DOWN:                               E->crs_curr.cy++;   break;
  }
}




void editorScroll(Editor_t * E) {

//  if (E->crs_curr.cy < E->lineCount) {
  //  editorUpdateRow( E, &E->row[E->crs_curr.cy]); // translate cursor x and y
//  } else {
    E->rx = E->crs_curr.cx;
 // }

  if (E->crs_curr.cy < E->crs_curr.rowoff) {
    E->crs_curr.rowoff = E->crs_curr.cy;
  }
  
  if (E->crs_curr.cy >= E->crs_curr.rowoff + E->windowrows) {
    E->crs_curr.rowoff = E->crs_curr.cy - E->windowrows + 1;
  }
  
  
  if (E->rx < E->crs_curr.coloff) {
    E->crs_curr.coloff = E->rx;
  }
  if (E->rx >= E->crs_curr.coloff + E->windowcols) {
    E->crs_curr.coloff = E->rx - E->windowcols + 1;
  }
  
}







/*
char * editorAttribToESC(int a) {
  switch (a) {
    case HL_DIGIT:     return ANSI_RedF;
    case HL_MATCH:     return ANSI_LTGreyF;
    case HL_STRING:    return ANSI_MagentaF;
    case HL_COMMENT:   return ANSI_CyanF;
    case HL_KEYWORD:   return ANSI_GreenF;
    case HL_TYPE:      return ANSI_YellowF;
    case HL_DIRECTIVE: return ANSI_BlueF;
    
    case HL_RB1:       return ANSI_BrownF;
    case HL_RB2:       return ANSI_RedF;
    case HL_RB3:       return ANSI_OrangeF;
    case HL_RB4:       return ANSI_YellowF;
    case HL_RB5:       return ANSI_GreenF;
    case HL_RB6:       return ANSI_BlueF;
    case HL_RB7:       return ANSI_PurpleF;
    case HL_RB8:       return ANSI_LTGreyF;    
        
 //   default:           return ANSI_ColorRST;
    default:           return ANSI_BrtBlueF;
  }
}
*/

/*
// per line syntax hilighting.
char * genAttribs(Editor_t * E, erow *row) {
   char          *tmpAttr;             // prehilighting
   char          *toks;                 // tokenizing copy
   char          *tok;
   char          *ptok;                 // padded token
   char          *p;                   // char pointer for walking the input
   char           src = 0, rrc = 0;
   static char    crc = 0;                // rainbow counters!
   int            x;                   // horizontal tracking
      
  tmpAttr     = strdup(row->chars);                 // tempate out as large as the file data for line
  memset(tmpAttr, HL_NORMAL, strlen(row->chars));   // default to normal hilight.
                
  toks        = strdup(row->chars);                 // copy for tokenizing   
      
  // whole word hilighting
  tok = strtok(toks, SEPCHARS);  
  while (tok != NULL) { 
     
    asprintf( &ptok, " %s ", tok); // make a padded version of the token       
    
    if (0) {
    } else if (strstr(KEYWORDS, ptok) != NULL) {
      memset(&tmpAttr[tok-toks], HL_KEYWORD, strlen(tok));            //  hilight word
    } else if (strstr(DIRECTIVES, ptok) != NULL) {
      memset(&tmpAttr[tok-toks], HL_DIRECTIVE, strlen(tok));          //  hilight word
    } else if (strstr(TYPES, ptok) != NULL) {
      memset(&tmpAttr[tok-toks], HL_TYPE, strlen(tok));               //  hilight word
    } 
                
    free(ptok);
    tok = strtok(NULL, SEPCHARS);
  }
       
  free(toks);  // at this point the result is in tmpAttr       
                  
  for (p = row->chars, tok = tmpAttr; *p; p++, x++, tok++ ) {          // character hilighting
  
    if (0) {
    } else if (*p == '(') {
      *tok = BRACKETRAINBOW[rrc];   rrc = (rrc+1)&7;
    } else if (*p == ')') {
      rrc = (rrc==0)?7:(rrc-1);  *tok = BRACKETRAINBOW[rrc];    
          
    } else if (*p == '[') {
      *tok = BRACKETRAINBOW[src];   src = (src+1)&7;
    } else if (*p == ']') {
      src = (src==0)?7:(src-1);  *tok = BRACKETRAINBOW[src];                        
      
    } else if (*p == '{') {
      *tok = BRACKETRAINBOW[crc];   crc = (crc+1)&7;
    } else if (*p == '}') {
      crc = (crc==0)?7:(crc-1);  *tok = BRACKETRAINBOW[crc];         
      
    } 
   }
   
   if ((toks = strstr(row->chars, "//")) != NULL) {
     memset(&tmpAttr[toks - row->chars], HL_COMMENT, strlen(row->chars) - (toks - row->chars) );
   }  
   
   //if ((toks = strstr(row->chars, E->dialog.el.buff)) != NULL)   memset(&tmpAttr[toks - row->chars], HL_MATCH, strlen(E->dialog.el.buff));            //  hilight word  
    
   return tmpAttr;
}
*/



void die(const char *s) {
  
  writes(ANSI_MouseOff);
  writes(ANSI_Cls);
  writes(ANSI_Home);
    
  perror(s);
  exit(1);
}
































