#ifndef __terminal_h
  #define __terminal_h

  #include <string.h>
  #include <stdio.h>
  #include <termios.h>
  #include <unistd.h>
  #include <sys/select.h>
  #include <stdlib.h>
  #include <errno.h>
  #include <sys/ioctl.h>
  
  
  #include "danStuff.h"
  #include "returnValues.h"


enum editorKey {
  BS_KEY      = 0x7F,
  KEY_NONE = 1000,
  
  KEY_LEFT, KEY_RIGHT, KEY_UP, KEY_DOWN,
  SHIFT_LEFT, SHIFT_RIGHT, SHIFT_UP, SHIFT_DOWN,
  
  PAGE_UP, PAGE_DOWN,
  
  KEY_HOME, KEY_END, SHIFT_HOME, SHIFT_END,
  
  KEY_DEL,
  
  MOUSE_EVENT, TIMER_EVENT, KEY_EVENT,
  SUBMIT_EVENT, CANCEL_EVENT,
  
  CTRL_A, CTRL_Q, CTRL_F, CTRL_S, CTRL_N, CTRL_O, CTRL_L, 
  
  ALT_0, ALT_1, ALT_2, ALT_3, ALT_4,
  
  KEY_F3,
  
  KEY_SHIFT_F3,
  
  FIND_RESET,
  END_LIST_KEY
  
};


typedef struct mouseEvent_s {  // mouse data carrier
  unsigned char buttons;
  unsigned char X;
  unsigned char Y;
} mouseEvent_t;


mouseEvent_t M;  // back in my day, 1 letter was all we had for variable names.


typedef struct Console_s {  
  int screenrows;
  int screencols;
  struct termios orig_termios;
} Console_t;

Console_t    C;


  #define  ANSI_Cls       "\e[2J"
  #define  ANSI_Home      "\e[H"
  #define  ANSI_Hide      "\e[?25l"
  #define  ANSI_Show      "\e[?25h"
  #define  ANSI_ClrLine   "\e[K"
  #define  ANSI_INV       "\e[7m"
  #define  ANSI_NORM      "\e[m"

  // mode 2 mouse events
  #define  ANSI_MouseOn   "\e[?1000h"
  #define  ANSI_MouseOff  "\e[?1000l"

  // another fine grey is 8
  #define  ANSI_GreyF     "\e[38;5;237m"
  #define  ANSI_RedF      "\e[31m"
  #define  ANSI_YellowF   "\e[33m"
  #define  ANSI_BlueF     "\e[34m"
  #define  ANSI_BrtBlueF  "\e[38;5;19m"
  #define  ANSI_MagentaF  "\e[35m"
  #define  ANSI_CyanF     "\e[36m"
  #define  ANSI_GreenF    "\e[32m"
  #define  ANSI_BrownF    "\e[38;5;52m"
  #define  ANSI_OrangeF   "\e[38;5;166m"
  #define  ANSI_PurpleF   "\e[38;5;54m"
  #define  ANSI_LTGreyF   "\e[38;5;245m"
  #define  ANSI_ColorRST  "\e[0m"

  #define CTRL_KEY(k) ((k) & 0x1f)


  // console IF
  Status_t     enableRawMode        ( ) ;
  Status_t     disableRawMode       ( ) ;
  int          checkStdin           ( void );
  void         writes               ( char * s);
  int          getWindowSize        ( int *rows, int *cols);
  int          escKeyLookup         ( char *s );
  unsigned int extCharLookup        ( unsigned char c );
  int          stdinFSM             ( char c) ;


#endif
